---
title: Modin
author: Steve Kossouho
---

# Accélération de la manipulation de données

----

## Qu'est-ce que Modin ?

Modin est une bibliothèque Python conçue pour offrir une alternative plus rapide et plus efficace à la bibliothèque pandas en exploitant pleinement le multicœur. Elle permet de travailler avec des DataFrames volumineux sans avoir à changer de syntaxe, car elle utilise la même API que pandas.

Avec Modin, les opérations sur les grands ensembles de données peuvent s’exécuter plus rapidement, en se répartissant sur plusieurs cœurs de processeur. L'objectif de Modin est d'optimiser le traitement des données en minimisant le code à modifier pour obtenir des gains de performance significatifs.

![Logo de Modin](assets/images/modin-logo.png)

----

## Pourquoi utiliser Modin ?

Modin propose des améliorations notables en termes de performance par rapport à pandas, surtout lorsque les opérations de traitement des données deviennent lourdes. Cette bibliothèque permet d'effectuer des opérations de manière distribuée sans effort supplémentaire, tout en maintenant une syntaxe familière.

Les situations idéales pour Modin incluent :

- Traitement de gros volumes de données
- Opérations lourdes en ressources de calcul
- Besoin de parallélisme sans passer à des outils plus complexes comme Spark

[Documentation de Modin](https://modin.readthedocs.io/)

----

## Installer Modin

Pour installer Modin, exécutez la commande suivante dans un terminal :

```bash {.numberLines}
pip install modin[ray]   # Pour un backend Ray
pip install modin[dask]  # Pour un backend Dask
```

Modin nécessite soit Ray, soit Dask comme backend pour paralléliser les opérations.

----

## Comment fonctionne Modin ?

Modin s'utilise comme pandas, en permettant de manipuler des DataFrames en parallèle, ce qui accélère de nombreuses opérations standards. Vous n'avez pas besoin de réapprendre de nouvelles méthodes ; la plupart des méthodes pandas sont supportées dans Modin.

Modin fonctionne comme un **wrapper** autour de pandas : il distribue les calculs en plusieurs partitions, chacune traitée sur un noyau CPU distinct.

----

### Exemple simple

Voici comment utiliser Modin pour créer et manipuler un DataFrame. L’API est similaire à celle de pandas :

```python {.numberLines}
import modin.pandas as pd

# Chargement des données
df = pd.read_csv("large_dataset.csv")

# Manipulation des données
df["new_column"] = df["existing_column"] * 2
df.head()
```

Ce code s’exécuterait plus rapidement qu’avec pandas si le jeu de données est volumineux, grâce à la parallélisation des tâches.

----

### Chargement de données volumineuses

L'une des fonctionnalités les plus utiles de Modin est sa capacité à lire des fichiers CSV massifs beaucoup plus rapidement que pandas :

```python {.numberLines}
import modin.pandas as pd

df = pd.read_csv("large_dataset.csv")
print(df.shape)
```

Modin utilise une technique de **partitionnement** pour diviser le fichier en morceaux et les traiter simultanément.

----

## Opérations de transformation

Modin prend en charge la plupart des opérations Pandas standard comme `groupby`{.py}, `merge`{.py}, et `join`{.python}. Ces opérations sont distribuées, ce qui permet de les réaliser rapidement.

```python {.numberLines}
# GroupBy avec Modin
grouped_df = df.groupby("column_name").mean()
```

Modin optimise l’opération de groupement pour les jeux de données volumineux en utilisant le multicœur.

----

### Comparaison de performances

Voici une comparaison illustrative des performances entre pandas et Modin (à tester dans Jupyter) :

```python {.numberLines}
import pandas as pd
import modin.pandas as mpd

# Avec pandas
df = pd.read_csv("large_dataset.csv")
%time df.groupby("column_name").mean()

# Avec Modin
mdf = mpd.read_csv("large_dataset.csv")
%time mdf.groupby("column_name").mean()
```

Modin affiche un temps de calcul inférieur en utilisant plusieurs cœurs. Cela ne se verra que lorsque vous travaillez avec des jeux de données volumineux.

----

### Compatibilité avec pandas

Modin est conçu pour fonctionner de manière interchangeable avec pandas, de sorte que la plupart des fonctions pandas courantes sont compatibles avec Modin. Cela permet une **transition facile** pour ceux qui utilisent déjà pandas pour le traitement des données.

Cependant, il peut y avoir des différences mineures en termes de prise en charge complète de certaines méthodes pandas moins fréquemment utilisées.

----

## Travailler avec des backends différents

Modin peut être configuré pour utiliser différents backends de calcul tels que **Ray** ou **Dask**. Ces deux bibliothèques sont des frameworks de calcul distribués bien connus, chacun avec ses particularités.

Pour choisir un backend, vous devez spécifier l'option correspondante lors de l'installation de Modin :

```bash {.numberLines}
pip install modin[ray]
# ou
pip install modin[dask]
```

Ray est souvent préféré pour les opérations nécessitant une faible latence, tandis que Dask est apprécié pour sa flexibilité dans le traitement des gros volumes de données.

----

### Exemple de code utilisant Ray ou Dask

```python {.numberLines}
# Initialisation avec Ray
import modin.config as cfg
cfg.Engine.put("ray")

import modin.pandas as pd
df = pd.read_csv("large_dataset.csv")
```

Modin détecte automatiquement le backend configuré et exécute les opérations de manière adaptée.

----

## Limitations et cas particuliers

Bien que Modin accélère de nombreuses opérations, certaines fonctionnalités de pandas ne sont pas encore entièrement prises en charge. Les opérations très complexes ou nécessitant un traitement unique peuvent ne pas bénéficier de la même accélération.

Quelques limitations à garder en tête :

- Certaines fonctions pandas avancées sont partiellement compatibles
- Modin est plus performant pour les opérations CPU-intensives

Pour des détails, reportez-vous à la [documentation Modin](https://modin.readthedocs.io/en/latest/supported_apis.html).
